@@ -240,6 +240,7 @@ class HomeConnect(DataClassJsonMixin):
         event_source = None
         while True:
             try:
+                _LOGGER.debug("Connecting to SSE stream")
                 event_source = await self._api.async_get_event_stream('/api/homeappliances/events')
                 await event_source.connect()
                 self.status |= self.HomeConnectStatus.UPDATES
@@ -285,6 +286,8 @@ class HomeConnect(DataClassJsonMixin):
                     await event_source.close()
                     event_source = None
 
+        _LOGGER.debug("Exiting SSE event stream")
+
 
     async def _async_process_updates(self, event:MessageEvent):
         """ Handle the different kinds of events received over the SSE channel """
